#!/usr/bin/env python
# SPDX-License-Identifier: GPL-3.0-or-later

from base64 import b64encode
from configparser import ConfigParser
from datetime import datetime
from http.server import BaseHTTPRequestHandler, HTTPServer
from io import BytesIO
from re import findall, search
from urllib.parse import parse_qs, urlparse

from bs4 import BeautifulSoup
from mariadb import connect
from matplotlib import pyplot
from matplotlib.dates import DateFormatter
from requests import get

config_daemon = ConfigParser()
config_daemon.read("/run/credentials/climated.service/conf")
hostname = config_daemon.get("server", "hostname")
port = config_daemon.getint("server", "port")
tokenlist = config_daemon.get("server", "tokenlist").split(",")
db_user = config_daemon.get("db", "user")
db_password = config_daemon.get("db", "password")
db = config_daemon.get("db", "db")
measurement_temperature = config_daemon.get("measurement", "temperature")
measurement_relative_humidity = config_daemon.get("measurement", "relative_humidity")
measurement_co2 = config_daemon.get("measurement", "co2")

DB_QUERY_LAST = "SELECT * FROM `history` WHERE `itemid` = {} ORDER BY `clock` DESC LIMIT 1;"
DB_QUERY_INTERVAL = "SELECT * FROM `history` WHERE `itemid` = {} AND `clock` > UNIX_TIMESTAMP(CURRENT_TIMESTAMP - INTERVAL {}) ORDER BY `clock` ASC;"
PRESSURE_NORMAL = 1013.25
URL_GRAPH_ZAB = "http://portal.chmi.cz/files/portal/docs/poboc/PR/grafy/br/B2BZAB01.PNG"
URL_GRAPH_TUR = "http://portal.chmi.cz/files/portal/docs/poboc/OS/KW/Captor/tmp/DMULTI-B2BTUR01.gif"
URL_TABLE_TUR = (
    "http://portal.chmi.cz/files/portal/docs/meteo/opss/pocasicko_nejnovejsi/st_11723_en.html"
)
URL_BRNO_METEOGRAMS_META = "http://portal.chmi.cz/files/portal/docs/meteo/ov/aladin/results/public/meteogramy/data/mdirs.txt"
URL_BRNO_METEOGRAMS = "http://portal.chmi.cz/files/portal/docs/meteo/ov/aladin/results/public/meteogramy/data/{}/42.png"
URL_SK_METEOGRAMS = "https://www.shmu.sk/data/datanwp/v2/meteogram/al-meteogram_{}-{}-{}-nwp-.png"
URL_TXT_HVEZDARNA_BRNO = "https://www.hvezdarna.cz/meteo/lastmeteodata.php"
URL_GRAPH_HVEZDARNA_BRNO = (
    "https://www.hvezdarna.cz/meteo/grafy/graf-{}.php?datum={}&typ=denni&stat="
)
URL_CAMERA_PAGE_HVEZDARNA_BRNO = "https://www.hvezdarna.cz/zive/kamera/"
URL_CAMERA_DAY_PICTURE_HVEZDARNA_BRNO = "https://www.hvezdarna.cz/ryba/day_latest.jpg?{}"
URL_CAMERA_NIGHT_PICTURE_HVEZDARNA_BRNO = "https://www.hvezdarna.cz/ryba/night_latest.png?{}"
URL_HVEZDARNA_BRNO = "https://www.hvezdarna.cz"
URL_GRAPH_CO2_HVEZDARNA_BRNO = (
    "https://www.hvezdarna.cz/czechglobe/zeremevesmir/grafy/graf-{}.php?datum={}&typ=denni&stat="
)
SHMU_ID_BRATISLAVA = "32737"
SHMU_ID_NOVE_ZAMKY = "31412"


hvezdarna_bits = ["teplota", "tlak", "vlhkost", "vitr"]
co2_bits = ["koncentraceCO2", "tokCO2", "rychlostV", "smerV"]


def fetch_last_value(measurement):
    cnx = connect(user=db_user, password=db_password, database=db)
    cursor = cnx.cursor()

    cursor.execute(DB_QUERY_LAST.format(measurement))
    result = cursor.fetchall()
    time = result[0][1]
    value = result[0][2]

    cursor.close()
    cnx.close()

    return time, value


def room_now():
    time1, temp = fetch_last_value(measurement_temperature)
    _, humidity = fetch_last_value(measurement_relative_humidity)
    _, co2 = fetch_last_value(measurement_co2)

    temperature = round(temp, 2)
    if co2 < 800:
        co2_level = "Normal"
    elif 800 <= co2 < 1200:
        co2_level = "High"
    else:
        co2_level = "Extreme"
    timestamp = datetime.fromtimestamp(time1).strftime("%d.%m.%Y %H:%M:%S")

    return f"""<div class="meteodata"><p><b>Time:</b> {timestamp}<br/>
<b>Temperature:</b> {temperature} °C<br/>
<b>Humidity:</b> {humidity} %<br/>
<b>CO₂:</b> {co2} ppm (<i>{co2_level}</i>)</p></div>"""


def graphs_get_data(cursor, query, measurement, interval):
    points_x = []
    points_y = []

    cursor.execute(query.format(measurement, interval))
    result = cursor.fetchall()
    for i in result:
        points_x.append(datetime.fromtimestamp(i[1]))
        points_y.append(i[2])

    return points_x, points_y


def graphs_plot(points_x, points_y, title, units, color):
    fig = pyplot.figure()

    pyplot.plot(points_x, points_y, color=color)
    pyplot.title(title)
    pyplot.ylabel(units)
    pyplot.gcf().autofmt_xdate()
    x_axis_format = DateFormatter("%H:%M")
    pyplot.gca().xaxis.set_major_formatter(x_axis_format)
    pyplot.grid(True)

    buffer = BytesIO()
    fig.savefig(buffer, format="png")
    buffer.seek(0)

    pyplot.close(fig)

    return buffer


def graphs_common(what, whatid, unit, color, interval):
    cnx = connect(user=db_user, password=db_password, database=db)
    cursor = cnx.cursor()

    points_x, points_y = graphs_get_data(cursor, DB_QUERY_INTERVAL, whatid, interval)

    cursor.close()
    cnx.close()

    return b64encode(graphs_plot(points_x, points_y, what, unit, color).getvalue()).decode("utf-8")


def img(mime, data):
    return f'<p class="graph"><img class="graph" src="data:image/{mime};base64,{data}"/></p>'


def room_common(interval):
    charts = ""
    charts += img(
        "png", graphs_common("Temperature", measurement_temperature, "°C", "red", interval)
    )
    charts += img(
        "png", graphs_common("Humidity", measurement_relative_humidity, "%", "blue", interval)
    )
    charts += img("png", graphs_common("CO₂", measurement_co2, "ppm", "green", interval))

    return charts


def room_6h():
    return room_common("6 HOUR")


def room_24h():
    return room_common("1 DAY")


def brno_now():
    page = get(URL_TABLE_TUR, timeout=10)
    if page.status_code == 200:
        soup = BeautifulSoup(page.content, features="html.parser")
        time = soup.body.find_all("table")[0].find_all("tr")[1].find_all("td")[0].text

        rows = soup.body.find_all("table")[1].find_all("tr")
        for i in rows:
            cols = i.find_all("td")
            if len(cols) < 3:
                continue
            description = cols[0].text
            value = cols[2].text
            if description.startswith("Temperature"):
                temperature = value
            elif description.startswith("Relative humidity"):
                humidity = value
            elif description.startswith("Air pressure at station"):
                pressure = value
                pressure_float = float(findall(r"\d+\.\d+", pressure)[0])
            elif description.startswith("Pressure reduced"):
                pressure_normalised = value
                pressure_pmn = round(
                    (float(findall(r"\d+\.\d+", pressure_normalised)[0]) - PRESSURE_NORMAL)
                    / PRESSURE_NORMAL
                    * 1000,
                    2,
                )
                if pressure_pmn > 0:
                    pressure_pmn_sign = "+"
                elif pressure_pmn < 0:
                    pressure_pmn_sign = "-"
                else:
                    pressure_pmn_sign = ""
                pressure_pmn = abs(pressure_pmn)
            elif description.startswith("Pressure trend"):
                pressure_trend = value
                pressure_trend_pmn = round(
                    float(findall(r"\-?\d+\.\d+", pressure_trend)[0]) / pressure_float * 1000, 2
                )
                if pressure_trend_pmn > 0:
                    pressure_trend_pmn_sign = "↑"
                elif pressure_trend_pmn < 0:
                    pressure_trend_pmn_sign = "↓"
                else:
                    pressure_trend_pmn_sign = ""
                pressure_trend_pmn = abs(pressure_trend_pmn)
            elif description.startswith("Wind speed"):
                wind_speed = value
            elif description.startswith("Wind direction"):
                wind_direction = value
            elif description.startswith("Cloudiness"):
                clouds = value
            elif description.startswith("Cumulative precipitation (hourly)"):
                precipitation = value
            elif description.startswith("Horizontal range of vision"):
                visibility = value
        msg = ""
        msg += '<h1 class="header">Tuřany</h1>'
        msg += f"""<div class="meteodata"><p><b>Time</b>: {time}<br/>
<b>Temperature:</b> {temperature}<br/>
<b>Humidity:</b> {humidity}<br/>
<b>Air pressure:</b> {pressure} (<i>{pressure_pmn_sign}{pressure_pmn} ‰, {pressure_trend_pmn_sign}{pressure_trend_pmn} ‰</i>)<br/>
<b>Wind:</b> {wind_speed} (<i>{wind_direction}</i>)<br/>
<b>Clouds:</b> {clouds}<br/>
<b>Precipitation (<i>1 hour</i>):</b> {precipitation}<br/>
<b>Visibility:</b> {visibility}</p></div>"""
    else:
        msg += "<p>Something went wrong while processing data from ČHMÚ Tuřany page.</p>"

    page = get(URL_GRAPH_TUR, timeout=10, stream=True)
    if page.status_code == 200:
        msg += img("gif", b64encode(page.raw.read()).decode("utf-8"))
    else:
        msg += "<p>Something went wrong while processing graph from ČHMÚ Tuřany page.</p>"

    msg += '<h1 class="header">Žabovřesky</h1>'

    page = get(URL_GRAPH_ZAB, timeout=10, stream=True)
    if page.status_code == 200:
        msg += img("png", b64encode(page.raw.read()).decode("utf-8"))
    else:
        msg += "<p>Something went wrong while processing graph from ČHMÚ Žabovřesky page.</p>"

    datestamp = None
    page = get(URL_TXT_HVEZDARNA_BRNO, timeout=10, stream=True)
    if page.status_code == 200:
        content = page.content.decode("utf-8")
        _, datestamp, timestamp, temperture, humidity, wind, feelslike, pressure = content.split(
            " "
        )
        realdatetime = datetime.strptime(datestamp + " " + timestamp, "%Y-%m-%d %H:%M:%S")
        realdatetimestr = realdatetime.strftime("%d.%m.%Y %H:%M:%S")
        msg += '<h1 class="header">Hvězdarna</h1>'
        msg += f"""<div class="meteodata"><p><b>Time</b>: {realdatetimestr}<br/>
<b>Temperature:</b> {temperture} °C (<i>🌡 {feelslike} °C</i>)<br/>
<b>Humidity:</b> {humidity} %<br/>
<b>Air pressure:</b> {pressure} hPa<br/>
<b>Wind:</b> {wind} m/s</p></div>"""
    else:
        msg += "<p>Something went wrong while processing data from Hvězdarna Brno page.</p>"

    for i in hvezdarna_bits:
        graph = get(URL_GRAPH_HVEZDARNA_BRNO.format(i, datestamp), timeout=10, stream=True)
        if graph.status_code == 200:
            msg += img("png", b64encode(graph.raw.read()).decode("utf-8"))
        else:
            msg += f'<p>Something went wrong while processing Hvězdarna Brno "{i}" image.</p>'

    page = get(URL_CAMERA_PAGE_HVEZDARNA_BRNO, timeout=10)
    if page.status_code == 200:
        day_camera_timestamp = search(r"day_latest\.jpg\?([0-9]+)", page.content.decode())[1]
        day_camera_img = get(
            URL_CAMERA_DAY_PICTURE_HVEZDARNA_BRNO.format(day_camera_timestamp),
            timeout=10,
            stream=True,
        )
        if day_camera_img.status_code == 200:
            msg += img("jpeg", b64encode(day_camera_img.raw.read()).decode("utf-8"))
        else:
            msg += "<p>Something went wrong while processing Hvězdarna Brno day camera image.</p>"

        night_camera_timestamp = search(r"night_latest\.png\?([0-9]+)", page.content.decode())[1]
        night_camera_img = get(
            URL_CAMERA_NIGHT_PICTURE_HVEZDARNA_BRNO.format(night_camera_timestamp),
            timeout=10,
            stream=True,
        )
        if night_camera_img.status_code == 200:
            msg += img("png", b64encode(night_camera_img.raw.read()).decode("utf-8"))
        else:
            msg += "<p>Something went wrong while processing Hvězdarna Brno night camera image.</p>"
    else:
        msg += "<p>Something went wrong while processing Hvězdarna Brno camera page.</p>"

    return msg


def brno_forecast():
    msg = ""

    page = get(URL_BRNO_METEOGRAMS_META, timeout=10)
    if page.status_code == 200:
        content = page.content.decode("utf-8")
        newest = content.partition("\n")[0]
        meteogram = get(URL_BRNO_METEOGRAMS.format(newest), timeout=10, stream=True)
        if meteogram.status_code == 200:
            msg += img("png", b64encode(meteogram.raw.read()).decode("utf-8"))
        else:
            msg += "<p>Something went wrong while processing ČHMÚ meteogram image.</p>"
    else:
        msg += "Something went wrong while processing ČHMÚ meteogram metadata.</p>"

    return msg


def brno_co2():
    msg = ""

    datestamp = None
    page = get(URL_TXT_HVEZDARNA_BRNO, timeout=10, stream=True)
    if page.status_code == 200:
        content = page.content.decode("utf-8")
        _, datestamp, _, _, _, _, _, _ = content.split(" ")
    else:
        msg += "<p>Something went wrong while processing data from Hvězdarna Brno page.</p>"

    if datestamp:
        for i in co2_bits:
            graph = get(URL_GRAPH_CO2_HVEZDARNA_BRNO.format(i, datestamp), timeout=10, stream=True)
            if graph.status_code == 200:
                msg += img("png", b64encode(graph.raw.read()).decode("utf-8"))
            else:
                msg += (
                    '<p>Something went wrong while processing Hvězdarna Brno CO₂ "{i}" image.</p>'
                )

    return msg


def slovakia_forecast():
    now = datetime.utcnow()
    hour = int(now.strftime("%H"))

    shmu_date = None
    shmu_time = None
    if 0 <= hour < 6:
        shmu_date = (now - datetime.timedelta(days=1)).strftime("%Y%m%d")
        shmu_time = "1800"
    elif 6 <= hour < 12:
        shmu_date = now.strftime("%Y%m%d")
        shmu_time = "0000"
    elif 12 <= hour < 17:
        shmu_date = now.strftime("%Y%m%d")
        shmu_time = "0600"
    elif 17 <= hour <= 23:
        shmu_date = now.strftime("%Y%m%d")
        shmu_time = "1200"

    msg = ""

    meteogram = get(
        URL_SK_METEOGRAMS.format(SHMU_ID_BRATISLAVA, shmu_date, shmu_time), timeout=10, stream=True
    )
    if meteogram.status_code == 200:
        msg += img("png", b64encode(meteogram.raw.read()).decode("utf-8"))
    else:
        msg += "<p>Something went wrong while processing SHMÚ Bratislava meteogram image.</p>"

    meteogram = get(
        URL_SK_METEOGRAMS.format(SHMU_ID_NOVE_ZAMKY, shmu_date, shmu_time), timeout=10, stream=True
    )
    if meteogram.status_code == 200:
        msg += img("png", b64encode(meteogram.raw.read()).decode("utf-8"))
    else:
        msg += "<p>Something went wrong while processing SHMÚ Nové Zámky meteogram image.</p>"

    return msg


class ClimateHandler(BaseHTTPRequestHandler):
    def __init__(self, request, client_address, server):
        super().__init__(request, client_address, server)
        self.token = None

    def send_code(self, code):
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def send_redirect(self, where):
        self.send_response(301)
        self.send_header("Location", f"{where}?token={self.token}")
        self.end_headers()

    def send_message(self, message):
        self.wfile.write(bytes(message, "utf-8"))

    def send_title(self, title):
        meta = """<meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
<meta name="viewport" content="width=device-width, initial-scale=1">
<meta name="description" content="Check room climate and weather in .cz and .sk" />"""

        favicon = """<link rel="apple-touch-icon" sizes="180x180" href="/apple-touch-icon.png">
<link rel="icon" type="image/png" sizes="32x32" href="/favicon-32x32.png">
<link rel="icon" type="image/png" sizes="16x16" href="/favicon-16x16.png">
<link rel="manifest" href="/site.webmanifest">
<link rel="mask-icon" href="/safari-pinned-tab.svg" color="#5bbad5">
<meta name="msapplication-TileColor" content="#da532c">
<meta name="theme-color" content="#ffffff">"""

        stylesheet = """<link rel="stylesheet" href="climated.css">"""

        self.send_message(
            f'<!doctype html><html lang="en"><head>{meta}<title>{title}</title>{favicon}{stylesheet}</head><body>'
        )

    def a(self, href, selector, title):
        aclass = "selected" if selector == href else "none"

        return f'<a class="{aclass}" href="/{href}?token={self.token}">{title}</a>'

    def send_menu(self, selector):
        msg = f"""<p><table class="menu">
    <tr>
        <td class="menuitem">{self.a("room_now", selector, "Room now")}</td>
        <td class="menuitem">{self.a("room_6h", selector, "Room 6h")}</td>
        <td class="menuitem">{self.a("room_24h", selector, "Room 24h")}</td>
    </tr>
    <tr>
        <td class="menuitem">{self.a("brno_now", selector, "Brno now")}</td>
        <td class="menuitem">{self.a("brno_forecast", selector, "Brno forecast")}</td>
        <td class="menuitem">{self.a("brno_co2", selector, "Brno CO₂")}</td>
    </tr>
    <tr>
        <td class="menuitem" colspan="3">{self.a("slovakia_forecast", selector, "Slovakia forecast")}</td>
    </tr>
</table></p>"""
        self.send_message(msg)

    def send_epilogue(self):
        self.send_message("</body></html>")

    def do_GET(self):
        path = urlparse(self.path)
        query = parse_qs(path.query)
        try:
            self.token = query["token"][0]
            allowed = self.token in tokenlist
        except KeyError:
            allowed = False

        if not allowed:
            self.send_code(401)
            self.send_title("Access denied")
            self.send_message("<p>Unauthorised</p>")
            self.send_epilogue()
            return

        match path.path:
            case "/room_now":
                self.send_code(200)
                self.send_title("Room now")
                self.send_menu("room_now")
                self.send_message(room_now())
                self.send_epilogue()
            case "/room_6h":
                self.send_code(200)
                self.send_title("Room 6h")
                self.send_menu("room_6h")
                self.send_message(room_6h())
                self.send_epilogue()
            case "/room_24h":
                self.send_code(200)
                self.send_title("Room 24h")
                self.send_menu("room_24h")
                self.send_message(room_24h())
                self.send_epilogue()
            case "/brno_now":
                self.send_code(200)
                self.send_title("Brno now")
                self.send_menu("brno_now")
                self.send_message(brno_now())
                self.send_epilogue()
            case "/brno_forecast":
                self.send_code(200)
                self.send_title("Brno forecast")
                self.send_menu("brno_forecast")
                self.send_message(brno_forecast())
                self.send_epilogue()
            case "/brno_co2":
                self.send_code(200)
                self.send_title("Brno CO₂")
                self.send_menu("brno_co2")
                self.send_message(brno_co2())
                self.send_epilogue()
            case "/slovakia_forecast":
                self.send_code(200)
                self.send_title("Slovakia forecast")
                self.send_menu("slovakia_forecast")
                self.send_message(slovakia_forecast())
                self.send_epilogue()
            case _:
                self.send_redirect("/room_now")


def launch():
    server = HTTPServer((hostname, port), ClimateHandler)

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass

    server.server_close()


if __name__ == "__main__":
    launch()
